#  Copyright 2019, 1533 Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

try:
    from _driveline_video import VideoDecoder
except ImportError:
    raise ImportError('driveline-video is not installed')


class VideoHandler(object):

    def __init__(self, frame_handler=None):
        self.decoder = VideoDecoder()
        self.frame_handler = frame_handler

    def __call__(self, record, *args, **kwargs):
        try:
            self.decoder.decode(record.raw_record)
            while True:
                frame = self.decoder.next_frame()
                if frame is None:
                    break
                self.on_frame(frame)
        except:
            pass

    def on_frame(self, frame):
        self.frame_handler(frame)


__all__ = [
    'VideoHandler'
]
